#include <u.h>
#include <libc.h>
#include <bio.h>

#define bits(buf, off, len)	(((buf) >> (off)) & ((1 << (len)) -1))

enum {
	TShdrlen = 4,			/* TS frame header len */
	TSfrmlen = 188,			/* TS frame length */
	PIDempty = 0x1fff,		/* unused PID */
	Npid = 4096
};

int Pktsz = 188;			/* Mpeg frame size */
int verbose;
int debug;

typedef struct MpgTs MpgTs;
struct MpgTs {
	int fd;					/* input pipe */
	int out;				/* output pipe */
	uchar *buf;				/* packet buffer */
	uchar *pos;				/* current position in buffer */
	uchar *end;				/* pointer to last byte of buf */
	int audio;				/* audio PID chosen */
	int meta;				/* metadata PID chosen */
	int id;					/* stream ID chosen */
	int seq[Npid];			/* TS frame sequence number */
	int eslen;				/* length of elementry stream message */
};

int is_adts(MpgTs *ts);
int is_mpeg2(MpgTs *ts);

struct {						/* heuristic tests for ES stream contents */
	int (*valid)(MpgTs *ts);	/* validator */
	char *path;					/* path to codec */
	char *name;					/* what to print */
} Codecs[] = {
	{ is_adts,	"/bin/audio/adtsdec",	"aac" },
	{ is_mpeg2,	"/bin/audio/mad", 		"mp3" },
//	{ is_ac3,	"/bin/audio/ac3dec", 	"ac3" },
};

static void
xd(uchar *buf, int len)
{
	int i, j;

	for(i = 0; i < len; i += j){
		fprint(2, "\n\t%02x:\t", i);
		for(j = 0; j < len && j < 16; j++)
			fprint(2, "%02x ", (uint)buf[i+j]);
		fprint(2, " \t");
		for(j = 0; j < len && j < 16; j++)
			if(buf[i+j] >= ' ')
				fprint(2, "%c", (char)buf[i+j]);
			else
				fprint(2, ".");
	}
	fprint(2, "\n\n");
}

int
rd(MpgTs *ts, int n)
{
	int i, buf;

	if(ts->pos+n >= ts->end)
		sysfatal("rd() packet under-run");

	buf = 0;
	for(i = 0; i < n; i++)
		buf = (buf << 8) | *ts->pos++;
	return buf;
}

void
skip(MpgTs *ts, int n)
{
	if(ts->pos+n > ts->end)
		sysfatal("skip() packet under-run (%d > %d)", n, (int)(ts->end-ts->pos));
	ts->pos += n;
}

void
cpy(MpgTs *ts, int len)
{
	if(ts->pos+len > ts->end)
		sysfatal("cpy(%d) packet under-run", len);

	if(write(ts->out, ts->pos, len) != len)
		sysfatal("write failed, %r");
	ts->pos += len;
}

int
srch(MpgTs *ts)
{
	int rc;

	do{
		rc = read(ts->fd, ts->buf, 1);
		if(rc == 0)
			return 1;
		if(rc == -1)
			return -1;
	}while(*ts->buf != 'G');

	rc = readn(ts->fd, ts->buf+1, Pktsz -1);
	if(rc == 0)
		return 1;
	if(rc != Pktsz -1)
		return -1;
	ts->pos = ts->buf;
	ts->end = ts->pos + Pktsz;
	return 0;
}


int
codec(char *path)
{
	char *arg[10];
	int i, narg, fd[2];

	narg = 0;
	arg[narg++] = path;
	arg[narg++] = "-s";
	arg[narg++] = "-v";
	arg[narg] = nil;

	if(pipe(fd) < 0){
		fprint(2, "pipe failed %r");
		return -1;
	}

	switch(fork()){
	case -1:
		fprint(2, "fork failed %r");
		close(fd[0]);
		close(fd[1]);
		return -1;
	case 0:

		dup(fd[0], 0);
		close(fd[0]);
		close(fd[1]);
		for(i = 3; i < 63; i++)
			close(i);
		exec(path, arg);
		sysfatal("exec %s failed %r", path);
	default:
		close(fd[0]);
		break;
	}
	return fd[1];
}

int
is_adts(MpgTs *ts)
{
	int i;
	uvlong buf;

	if(ts->pos+7 >= ts->end)
		sysfatal("adts() packet under-run");

	buf = 0;
	for(i = 0; i < 7; i++)
		buf = (buf << 8) | ts->pos[i];

	if(bits(buf, 44, 12) != 0xfff)		/* preamble */
		return 0;
	if(bits(buf, 41, 2) != 0)			/* layer */
		return 0;
	if(bits(buf, 38, 2) == 0)			/* profile */
		return 0;
	if(bits(buf, 34, 4) == 15) 			/* sample rate */
		return 0;
	if(bits(buf, 33, 1) != 0)			/* private flag (unused) */
		return 0;
	if(bits(buf, 29, 1) != 0)			/* origin */
		return 0;
	if(bits(buf, 28, 1) != 0)			/* home */
		return 0;
	if(bits(buf, 23, 3) >= 8)			/* channel spec */
		return 0;
	return 1;
}

int
is_mpeg2(MpgTs *ts)
{
	int i;
	uint buf;

	if(ts->pos+4 >= ts->end)
		sysfatal("mpeg2() packet under-run");

	buf = 0;
	for(i = 0; i < 4; i++)
		buf = (buf << 8) | ts->pos[i];

	if(bits(buf, 21, 11) != 0x7ff)		/* frame sync */
		return 0;
	if(bits(buf, 19, 2) == 1)			/* version */
		return 0;
	if(bits(buf, 17, 2) == 0)			/* layer */
		return 0;
	if(bits(buf, 12, 4) == 0x0f) 		/* bitrate */
		return 0;
	if(bits(buf, 10, 2) == 3)			/* sample rate */
		return 0;
	if(bits(buf, 0, 2) == 2)			/* emphasis */
		return 0;
	return 1;
}

int
pes_audio(MpgTs *ts, int first)
{
	uint buf;
	int n, i, id, sync, peslen, hdrlen;

	if(first){
		sync = rd(ts, 3);
		if(sync != 1)					/* bad sync byte for */
			return -1;

		id = rd(ts, 1);
		if(id < 0xc0 || id > 0xdf)		/* not an audio elementry stream */
			return -1;

		peslen = rd(ts, 2);
		if(peslen == 0)
			sysfatal("zero (any) length audio stream, illegal\n");

		buf = rd(ts, 2);
		if(bits(buf, 14, 2) != 2)		/* no marker bits */
			return -1;

		hdrlen = rd(ts, 1);
		ts->eslen = peslen - (2 + hdrlen);

		skip(ts, hdrlen);
		if(ts->id == -1){				/* stream id not set yet */
			ts->id = id;
			print("stream 0x%x\n", id);
		}
		else
		if(ts->id != id)				/* not the stream we have chosen */
			return -1;

		if(ts->out == -1){				/* start codec if none yet */
			for(i = 0; i < nelem(Codecs); i++)
				if(Codecs[i].valid(ts)){
					ts->out = codec(Codecs[i].path);
					print("codec %s\n", Codecs[i].name);
					break;
				}
		}
		if(ts->out == -1)
			sysfatal("audio format not recognised, %r");

	}

	if(ts->out == -1)					/* cannot say it works, as we have no codec yet */
		return -1;

	n = ts->end-ts->pos;

	if(n > ts->eslen)
		n = ts->eslen;
	ts->eslen -= n;

	if(n > 0)
		cpy(ts, n);

	return 0;
}

int
pes_meta(MpgTs *ts, int first)
{
	int sync, id, peslen, buf, hdrlen;
	
if(debug)
	xd(ts->buf, ts->end-ts->buf);

	if(first){
		sync = rd(ts, 3);
if(debug)
	fprint(2, "sync %d\n", sync);
		if(sync != 1)					/* bad sync byte for */
			return -1;

		id = rd(ts, 1);
if(debug)
	fprint(2, "id %d\n", id);
		if(id != 0xbd)						/* not a private stream */
			return -1;

		peslen = rd(ts, 2);
		if(peslen == 0)
			sysfatal("zero (any) length metadata stream, illegal\n");

		buf = rd(ts, 2);
//		if(bits(buf, 14, 2) != 2)		/* no marker bits */
//			return -1;

		hdrlen = rd(ts, 1);
		ts->eslen = peslen - (2 + hdrlen);

		skip(ts, hdrlen);

	}
	return 0;
}

void
chkseq(MpgTs *ts, int pid, int seq)
{
	if(ts->seq[pid] == -1)
		ts->seq[pid] = seq;

	if(seq != ts->seq[pid]){
		if(verbose)
			print("error 'pid=0x%x broken sequence, got=%d want=%d'\n", pid, seq, ts->seq[pid]);
		ts->seq[pid] = seq;
	}
	ts->seq[pid] = (seq+1)%16;
}

int
transp(MpgTs *ts)
{
	uint buf;
	int rc, err, first, pid, scramb;
	int payload, adapt, seq, adlen;
	
	while((rc = srch(ts)) == 0){
		buf = rd(ts, 4);

		err = bits(buf, 23, 1);
		first = bits(buf, 22, 1);
		pid = bits(buf, 8, 13);
		scramb = bits(buf, 6, 2);
		adapt = bits(buf, 5, 1);
		payload = bits(buf, 4, 1);
		seq = bits(buf, 0, 4);

		if(err != 0)							/* demodulation failed */
			continue;
		if(scramb != 0)							/* scrambled (encrypted) */
			continue;
		if(pid == PIDempty)						/* padding, no useful content */
			continue;
		if(! payload)							/* nothing of interest inside */
			continue;

		if(adapt){								/* step over any adaption data */
			adlen = rd(ts, 1);
			skip(ts, adlen);
		}


		if(ts->audio == -1 || pid == ts->audio){ /* our pid or none chosen */
			if(pid == ts->audio)
				chkseq(ts, pid, seq);
			if(pes_audio(ts, first) == 0)
				if(ts->audio == -1)			/* audio pid not chosen yet */
					ts->audio = pid;
		}
//		else
//		if(pes_meta(ts, first) == 0){
//			if(pid == ts->meta)
//				chkseq(ts, pid, seq);
//			if(ts->meta == -1)					/* meta pid not chosen yet */
//				ts->meta = pid;
//		}
		else
		if(debug > 1){
			fprint(2, "err		%d\n", err);
			fprint(2, "first	%d\n", first);
			fprint(2, "pid		%d\n", pid);
			fprint(2, "scramb	%d\n", scramb);
			fprint(2, "pid		%d\n", pid);
			fprint(2, "adapt	%d\n", adapt);
			fprint(2, "payload	%d\n", payload);
			fprint(2, "seq		%d\n", seq);
			xd(ts->buf, ts->end-ts->buf);
		}
	}
	return rc;
}

int
setprio(int pid, int prio)
{
	int fd;
	char buf[128];

	snprint(buf, sizeof(buf), "/proc/%d/ctl", pid);
	if((fd = open(buf, OWRITE)) == -1)
		return -1;
	fprint(fd, "pri %d\n", prio);
	close(fd);
	return 0;
}

static void
usage(void)
{
	fprint(2, "usage: %s [-s] [-v] [-o file] [-n pktsz] [file]\n", argv0);
	fprint(2, "	  -s        stream mode, compatability with codecs\n");
	fprint(2, "   -o file   output file, do not envoke decoder\n");
	fprint(2, "   -n pktsz  mpeg packet size (%d)\n", Pktsz);
	fprint(2, "	  -v        verbose\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	MpgTs *ts;
	int i, rc;
	char *ofile;

	ofile = nil;
	ARGBEGIN{
	case 'o':
		ofile = EARGF(usage());
		break;
	case 's':
		break;
	case 'd':
		debug++;
		break;
	case 'v':
		verbose++;
		break;
	case 'n':
		Pktsz = atoi(EARGF(usage()));
		break;
	default:
		usage();
		break;
	}ARGEND;

	ts = mallocz(sizeof(MpgTs) + Pktsz, 1);
	if(ts == nil)
		sysfatal("No memory, %r");
	ts->buf = (uchar *)&ts[1];

	ts->id = -1;
	ts->audio = -1;
	ts->meta = -1;
	ts->out = -1;
	for(i = 0; i < Npid; i++)
		ts->seq[i] = -1;

	setprio(getpid(), 16);

	if(ofile)
		if((ts->out = create(ofile, OWRITE, 0644)) == -1)
			sysfatal("%s cannot create file, %r", ofile);

	rc = 0;
	if(argc == 0){
		ts->fd = 0;
		rc = transp(ts);
	}
	else
		for(i = 0; i < argc; i++){
			if((ts->fd = open(argv[i], OREAD)) == -1)
				sysfatal("%s cannot open, %r", argv[i]);
			rc |= transp(ts);
			close(ts->fd);
		}
	if(rc < 0)
		sysfatal("demux failed, %r");
	exits(nil);
}
