#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include <xml.h>
#include "dat.h"
#include "fns.h"

/*
 http://bitcdn-vix1.bitmovin.net/content/hevc/dashevc-live-2s.mpd
 http://rdmedia.bbc.co.uk/dash/ondemand/testcard/1/client_manifest-audio.mpd
 http://stackoverflow.com/questions/21634622/number-of-segments-in-a-dash-mpd
*/

typedef struct Aux Aux;
struct Aux {
	Chunk **last;
	char *media;		/* for templated files */
	char *init;			/* for templated files */
	char *base;			/* base of URL */
	char *title;
	double secs;
	int bw;
};

char *
strval(Xml *xp, Elem *ep, char *path, char *attr)
{
	Attr *ap;

	if(path)
		if((ep = xml_find(xp, ep, path)) == nil){
			werrstr("dash: %s not found", path);
			return nil;
		}

	for(ap = ep->attrs; ap; ap = ap->next)
		if(strcmp(ap->name, attr) == 0)
			return ap->value;

	werrstr("dash: %s/%s not found", path, attr);
	return nil;
}

double
pt2sec(char *s)
{
	char c;
	double n, sec;

	if(s == nil)
		return -1;
	if(cistrncmp(s, "pt", 2) != 0)
		return -1;
	s += 2;

	sec = 0;
	while(*s){
		n = strtod(s, &s);
		if(*s){
			c = tolower(*s++);
			switch(c){
			case 's':
				sec += n;
				break;
			case 'm':
				sec += n*60;
				break;
			case 'h':
				sec += n*60*60;
				break;
			default:
				sysfatal("dash: pt2sec() '%c' unknown modifier", c);
			}
		}
	}
	return sec;
}

void
add(Playlist *pl, char *tail)
{
	Aux *a;
	Chunk *ch;

	a = pl->aux;
	ch = emalloc(sizeof(Chunk));
	ch->bw = a->bw;
	ch->secs = a->secs;
	ch->url = esmprint("%s/%s", a->base, tail);
	if(a->title)
		ch->title = estrdup(a->title);

	*a->last = ch;
	a->last = &ch->next;
}

int
templated(Playlist *pl, Xml *xp)
{
	Aux *a;
	char *t;
	Elem *ep;

	if((ep = xml_find(xp, xp->root, "MPD/Period/AdaptationSet/SegmentTemplate")) == nil)
		return -1;

	a = pl->aux;
	if((t = strval(xp, ep, nil, "media")) == nil)
		return -1;
	a->media = estrdup(t);

	if((t = strval(xp, ep, nil, "duration")) == nil)
		return -1;
	a->secs = atoi(t);

	if((t = strval(xp, ep, nil, "timescale")) != nil)
		a->secs /= atoi(t);

	if((t = strval(xp, ep, nil, "initialization")) == nil)
		return -1;
	a->init = estrdup(t);

	return 0;
}

/* chose the most apropriate (highest bandwidth) representation offered */
Elem *
chose(Playlist *pl, Xml *xp)
{
	int n;
	Aux *a;
	char *t;
	Elem *best, *ep;

 	a = pl->aux;
	if((best = xml_find(xp, xp->root, "MPD/Period/AdaptationSet/Representation")) == nil)
		return nil;

	for(ep = best; ep; ep = ep->next){
		if((t = strval(xp, ep, nil, "mimeType")) == nil)
			continue;
		if(strcmp(t, "video/mp4") != 0)
			continue;
		if((t = strval(xp, ep, nil, "bandwidth")) == nil)
			continue;
		n = atoi(t);
		if(n > a->bw){
			a->bw = n;
			best = ep;
		}
	}
	return best;
}

int
enumerated(Playlist *pl, Xml *xp)
{
	Aux *a;
	char *t;
	Elem *ep;

	a = pl->aux;
	if((ep = chose(pl, xp)) == nil)
		return -1;

	/* add codec initialisation segemet, if present */
	if((t = strval(xp, ep, "Representation/SegmentBase/Initialization", "sourceURL")) == nil)
		return -1;
	add(pl, t);		/* FIXME: is initialisation really zero seconds? */

	/* segement duration */
	if((t = strval(xp, ep, "Representation/SegmentList", "duration")) == nil)
		return -1;
	a->secs = atof(t);


	/* segement urls */
 	if((ep = xml_find(xp, ep, "Representation/SegmentList/SegmentURL")) == nil){
		werrstr("dash: no Representation/SegmentList/SegmentURL");
		return -1;
	}
	for(; ep; ep = ep->next){
		if((t = strval(xp, ep, nil, "media")) == nil)
			continue;
		add(pl, t);
	}
	return 0;
}

int
parse(Playlist *pl, Xml *xp)
{
	Aux *a;
	Elem *ep;

	a = pl->aux;
 	if((ep = xml_find(xp, xp->root, "MPD/ProgramInformation/Title")) != nil && ep->pcdata != nil)
		a->title = ep->pcdata;

	if(enumerated(pl, xp) == -1)
		if(templated(pl, xp) == -1)
			return -1;

	return 0;
}

Playlist *
dash(char *url, int fd)
{
	Aux *a;
	Xml *xp;
	Playlist *pl;

	pl = emalloc(sizeof(Playlist));
	pl->isaudio = 1;
	a = emalloc(sizeof(Aux));
	a->last = &pl->chunks;
	pl->aux = a;

	a->base = baseurl(url);
	if((xp = xml_parse(fd, 0xffff, Fcrushwhite)) == nil)
		fprint(2, "dash: xml parse failed\n");

	parse(pl, xp);
	xml_free(xp);

	/* do not free base or title, they are not malloc'ed */
	free(a->init);
	free(a->media);
	free(a);
	pl->aux = nil;

	return pl;
}
