#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include "dat.h"
#include "fns.h"

#include "/sys/src/cmd/tcs/8859.h"

enum {
	Maxmeta = 256 * 16,		/* 8bit length, measuring 16byte chunks */
};

/*
 * We take great steps here to parse mal-formed fields as
 * they are (sadly) rather common
 */

typedef struct Xlate Xlate;
struct Xlate {
	char *src;
	char *dst;
	char *ign;			/* server's default string, ignore this */
	int split;			/* check for "song - artist" format */
};

/*
 * Remember, webfs crushes hyphens out of header names when it generates
 * its virtual filesystem so we need to list header names without
 * hyphens here too.
 */
static Xlate Base[] = {
	{ "icyname",			"station", 		"", 0 },
	{ "icygenre",			"genre", 		"", 0 },

	{ "icename",			"station", 		"", 0 },
	{ "icegenre",			"genre", 		"", 0 },

	{ nil, nil },
};

static Xlate Stream[] = {
	{ "streamtitle",	"title", 			"",	1 },
	{ "streamurl",		nil, 	 			"",	1 },
	{ nil, nil },
};

static void
xd(char *buf, int len)
{
	int i, j;

	for(i = 0; i < len; i += j){
		fprint(2, "\n\t%02x:\t", i);
		for(j = 0; j < len && j < 16; j++)
			fprint(2, "%02x ", ((Rune)buf[i+j] & 0xff));
		fprint(2, " \t");
		for(j = 0; j < len && j < 16; j++)
			if(buf[i+j] >= ' ')
				fprint(2, "%c", (Rune)buf[i+j] );
			else
				fprint(2, "%c", '.');
	}
	fprint(2, "\n\n");
}


static char *
dequote(char *s)
{
	char *p;

	if(*s != '\'')
		return s;
	s++;
	if((p = strrchr(s, '\'')) != nil)
		*p = 0;
	return s;
}

/*
 * Trim leading and trailing whitespace, and convert to utf-8.
 *
 * Shoutcast uses LatinN metadata, Icecast uses utf-8,
 * others seem a bit random. Rather than explicitly selecting
 * the characterset depending on the server type we try to
 * make an educated guess.
 *
 * As we don't know which Latin translation table to use
 * we just pick latin1 and hope for the best. Shoutcast2
 * uses utf-8 so things should get better as people migrate.
 */
static char *
clean(char *buf)
{
	int n;
	Rune r;
	char *i, *o;
	static char tmp[Maxmeta];
	uchar bom[] = { 0xef, 0xbb, 0xbf };

	i = buf;
	o = tmp;

	/*
	 * Addicted radio, and maybe others, add a
	 * suprious utf8 BOM (byte order mark). uft8
	 * has no such concept, the clue is in the
	 * name - utf8, i.e. 8 bit, no byte order (sigh).
	 */
	if(memcmp(i, bom, sizeof(bom)) == 0)
		i += 3;

	while(isspace(*i))
		i++;

	while(*i){
		n = chartorune(&r, i);
		if(r == Runeerror)
			r = tab8859_2[*(uchar *)i];
		o += runetochar(o, &r);
		i += n;
	}

	*o-- = 0;
	while(o >= tmp && isspace(*o))
		*o-- = 0;

	return tmp;
}

static int
split(char *str, char **arr, int num, char *delim)
{
	char *p;
	int n, len;

	if(dbg("icy"))
		fprint(2, "icy.split: '%s'\n", str);

	if((p = strstr(str, delim)) == nil)
		return 0;

	n = 0;
	len = strlen(delim);
	while(p && n < num){
		*p = 0;
		arr[n++] = str;
		str = p + len;
		p = strstr(str, delim);
	}
	if(n < num)
		arr[n++] = str;

	if(dbg("icy"))
		fprint(2, "icy.split: -> n=%d\n", n);
	return n;
}

/*
 * expand streamtitle="" metadata which often contain
 * multiple fields.
 */
static void
expand(char *str)
{
	int n;
	char *a[8];

	if((n = split(str, a, nelem(a), " - ")) == 0)
		n = split(str, a, nelem(a), " + ");
			
	switch(n){
	default:
		print("show %q\n", clean(str));
		break;
	case 2:
		print("artist %q\n", clean(a[0]));
		print("title %q\n", clean(a[1]));
		break;
	case 3:
		print("show %q\n", clean(a[0]));
		print("artist %q\n", clean(a[1]));
		print("title %q\n", clean(a[2]));
		break;
	}
}

			
/*
 * Multi field lines have very poorly defined quoting rules.
 * Whitespace is quoted with single quotes, however there is
 * no quote escape character. We do our best by looking for "';"
 * which often works but we may parse incorrectly somtimes.
 */
static int
multi(char *buf, Xlate *x)
{
	int rc;
	char *nam, *e, *val;

	e = "";
	rc = 0;
	for(nam = buf; *nam; nam = e){
		if((val = strchr(nam, '=')) == nil)
			continue;
		*val++ = 0;

		if(*val == '\''){
			val++;
			if((e = strstr(val, "';")) != nil){		/* 'text';text */
				*e = 0;
				e += 2;
			}
			else
			if((e = strrchr(val, '\'')) != nil)		/* 'text' */
				*e = 0;
			else
				e = strchr(val, 0);					/* 'text (badly formatted) */
		}
		else{
			if((e = strchr(val, ';')) != nil)		/* text;text */
				*e++ = 0;
			else
				e = strchr(val, 0);					/* text */
		}

		for(x = x; x->src; x++)
			if(cistrcmp(nam, x->src) == 0){
				if(x->dst && (x->ign == nil || strcmp(val, x->ign) != 0)){
					if(x->split)
						expand(val);
					else
						print("%s %q\n", x->dst, clean(val));
					rc++;
				}
				break;
			}
		if(dbg("icy") && x->src == nil)
			fprint(2, "icy.field '%s=%s' unknown\n", nam, val);
	}
	return rc;
}

int
icyhdr(Sess *s, int conn)
{
	Xlate *x;
	int metaint;
	char val[Maxmeta];

	if(httphdr(s, conn, "icymetaint", val, sizeof(val)) < 0)
		return -1;

	metaint = atoi(val);
	if(dbg("icy")){
		fprint(2, "icy.metaint %d\n", metaint);

		if(metaint == 0)	/* none offered */
			print("title '<no metatdata>'\n");
		else				/* initial state */
			print("title '<empty>'\n");
	}
	for(x = Base; x->src; x++)
		if(httphdr(s, conn, x->src, val, sizeof(val)) == 0)
			if(x->ign == nil || strcmp(val, x->ign) != 0){
				if(x->dst)
					print("%s %q\n", x->dst, clean(dequote(val)));
			}

	return metaint;
}

int
icyinband(int fd)
{
	int sz;
	uchar c;
	char buf[Maxmeta];

	if(read(fd, &c, 1) != 1)
		return -1;
	if(c == 0)				/* no data this time */
		return 0;
	sz = c * 16;

	if(readn(fd, buf, sz) < sz)
		return -1;
	buf[sz] = 0;

	if(dbg("icydump"))
		xd(buf, sz);


	if(dbg("icy"))
		fprint(2, "icy.inband %d '%s'\n", sz, buf);

	multi(buf, Stream);
	return 0;
}
