#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include "dat.h"
#include "fns.h"

/*
 * Note: we ignore any information we can work out as it may
 * be (often is) wrong:
 *
 * - disconunity - we always check the ContentType HTTP header or, 
 *	 filename extension if none, to chose a codec. If it changes we
 *	 open a new codec. All codecs reset themselves if the input
 *	 changes significantly (bitrate, numchans etc).
 *
 * also beware spurions carriage returns and linefeeds, and possible
 * missing carriage return or linefeed on last line of the file.
 *
 * chunk list is ordered by bitrate, but the bitrates may be in
 * random order as the playlist is read on the wire.
 *
 * Depending on the timing of when the playlist is read it may contain
 * duplicate entries for chunks already played.
 * This is handled in play() in main.c
 */

typedef struct Aux Aux;
typedef struct Record Record;

struct Aux {
	Chunk *chunk;		/* chunk currently being built */
};

struct Record {
	char *name;
	int (*func)(Playlist *, char *);
};

static int
sequence(Playlist *pl, char *args)
{
	pl->seqbase = strtoull(args, nil, 10);
	pl->chunked = 1;
	return 0;
}

static int
byterange(Playlist *pl, char *args)
{
	Aux *a;
	vlong len;
	Chunk *ch;

	a = pl->aux;
	if(a->chunk == nil)
		a->chunk = emalloc(sizeof(Chunk));
	ch = a->chunk;

	len = strtoll(args, &args, 10);
	if(*args++ != '@')
		sysfatal("%q badly formatted byterange", args);
	ch->start = strtoll(args, nil, 10);
	ch->end = ch->start+len -1;			/* -1 due to gaps and telegraph poles */
	return 0;
}

static int
stream(Playlist *pl, char *args)
{
	Aux *a;
	int i, n;
	Chunk *ch;
	char *fld[16];

	a = pl->aux;
	if(a->chunk == nil)
		a->chunk = emalloc(sizeof(Chunk));
	ch = a->chunk;

	n = gettokens(args, fld, nelem(fld), ",");
	for(i = 0; i < n; i++)
		if(cistrncmp(fld[i], "BANDWIDTH=", 10) == 0)
			ch->bw = strtol(fld[i]+10, nil, 10);
	return 0;
}


static int
extinf(Playlist *pl, char *args)
{
	Aux *a;
	Chunk *ch;

	pl->isaudio = 1;

	a = pl->aux;
	if(a->chunk == nil)
		a->chunk = emalloc(sizeof(Chunk));
	ch = a->chunk;

	ch->secs = strtod(args, &args);
	if(*args == ',')
		args++;
	parsetitle(ch, skipws(args));
	return 0;
}

static int
listtype(Playlist *pl, char *args)
{
	if(cistrcmp(args, "vod") == 0)
		pl->live = 0;
	return 0;
}

static int
endlist(Playlist *pl, char *)
{
	pl->live = 0;
	return 0;
}

Record Records[] = {
	{ "#EXT-X-STREAM-INF:",			stream },
	{ "#EXT-X-MEDIA-SEQUENCE:",		sequence },
	{ "#EXTINF:",					extinf },
	{ "#EXT-X-PLAYLIST_TYPE:",		listtype },
	{ "#EXT-X-ENDLIST",				endlist },		/* yes, no trailing colon */
	{ "#EXT-X-BYTERANGE:",			byterange },
	{ nil,							nil },
};

static int
cmp(Chunk *a, Chunk *b)
{
	if(Narrowband){
		if(a->bw > b->bw)
			return 1;
	}
	else{
		if(a->bw > b->bw)
			return 1;
	}
	return 0;
}

Playlist *
m3u8(char *url, int fd)
{
	int n;
	Aux *a;
	Record *r;
	Biobuf bin;
	Playlist *pl;
 	Chunk *ch, *t, **prev;
	char buf[8192], *base, *rec;

	base = baseurl(url);
	Binit(&bin, fd, OREAD);

	pl = emalloc(sizeof(Playlist));
	pl->live = 1;			/* assume live until we know different */

	a = emalloc(sizeof(Aux));
	pl->aux = a;

	while(Bgetline(&bin, buf, sizeof(buf)) != -1){
		rec = skipws(buf);
		if(*rec == 0)							/* skip blank lines */
			continue;

		if(dbg("m3u8"))
			fprint(2, "m3u8 %q\n", rec);

		if(cistrncmp(rec, "#EXT", 4) == 0){		/* structured comment */
			for(r = Records; r->name; r++){
				n = strlen(r->name);
				if(strncmp(rec, r->name, n) == 0){
					r->func(pl, rec + n);
					break;
				}
			}
			continue;
		}

		if(*rec != '#'){						/* relative filename or a URL */
			if(a->chunk == nil)
				a->chunk = emalloc(sizeof(Chunk));
			ch = a->chunk;
			a->chunk = nil;

			ch->url = absurl(base, rec);

			/* sort in decreasing bandwidth order */
			prev = &pl->chunks;
			for(t = pl->chunks; t; t = t->next){
				if(cmp(ch, t))
					break;
				prev = &t->next;
			}
			ch->next = *prev;
			*prev = ch;
			continue;
		}
	}

	Bterm(&bin);
	free(base);

	return pl;
}

