#include <u.h>
#include <libc.h>
#include <bio.h>
#include "dat.h"
#include "fns.h"

enum {			/* player status */
	Error,		/* give up */
	Retry,		/* playlist exhausted, retry to see if it has changed */
	Finished,	/* playlist exhausted, audio on demand / podcast, now exit */
};

char *Outfile;
char *Debug;
int Narrowband;		/* chose low bitrate chunked audio if given a choice */
int Touch;			/* do not load audio, read playlists and attach to servers (populate DNS) */

typedef struct Mime Mime;
static struct Mime {
	char *type;
	char *ext;
	Playlist *(*func)(char *url, int fd);
	char *codec;
} Mimetypes[] = {
	{ "audio/x-scpls", 					"pls",		pls },
	{ "application/pls+xml", 			"pls",		pls },				/* vpr.net weirdness */
	{ "audio/x-mpegurl",				"m3u8",		m3u8 },				/* ascii */
	{ "audio/mpegurl",					"m3u8",		m3u8 },				/* http://www.rozhlas.cz/audio/download/jazz_high_ogg.m3u */
	{ "application/x-mpegurl",			"m3u8",		m3u8 },				/* utf8 */
	{ "application/vnd.apple.mpegurl",	"m3u8",		m3u8 },				/* HLS chunked */
	{ "application/dash+xml",			"mpd",		dash },				/* MPEG DASH chunked */

	{ "audio/aacp", 					"aac",		nil,		"/bin/audio/adtsdec" },	/* AAC in MP2 TS */
	{ "audio/aac", 						"aac",		nil, 		"/bin/audio/adtsdec" },
	{ "audio/3gpp",						"3gpp",		nil, 		"/bin/audio/adtsdec" },
	{ "audio/3gpp2",					"3gpp2",	nil, 		"/bin/audio/adtsdec" },

	{ "audio/mpeg", 					"mpeg",		nil,		"/bin/audio/mad" },		/* MP1, MP2, or MP3 */
	{ "audio/x-mpeg", 					"mpeg",		nil,		"/bin/audio/mad" },
	{ "audio/mp2",						"mp2",		nil,		"/bin/audio/mad" },
	{ "audio/mp3",						"mp3",		nil,		"/bin/audio/mad" },

	{ "audio/x-aiff",					"aiff",		nil,		"/bin/audio/aiffdec" },

	{ "audio/x-m4a",					"m4a",		nil,		"/bin/audio/m4adec" },	/* AAC or ALAC in an MP4 wrapper */
	{ "audio/mp4",						"m4a",		nil,		"/bin/audio/m4adec" },	/* AAC or ALAC in an MP4 wrapper */
	{ "audio/mp4a-latm",				"m4a",		nil,		"/bin/audio/m4adec" },

	{ "video/mp4",						"m4s",		nil,		"/bin/audio/m4adec" },	/* ISO Base Media / MPEG-4 Part 14 */

	{ "audio/x-wav",					"wav",		nil,		"/bin/audio/wavdec" },

	{ "audio/x-flac",					"flac",		nil,		"/bin/audio/flacdec" },
	{ "application/ogg",				"ogg",		nil,		"/bin/audio/flacdec" },

	{ "video/MP2T", 					"ts",		nil,		"/bin/audio/mpgtsdec" },	/* HLS */

//	{ "audio/x-aifc",					"aifc",		nil,		"/bin/audio/aiffdec" },		/* ??? */
//	{ "audio/x-caf",					"caf",		nil,		"/bin/audio/cafdec" },		/* ??? */

	{ nil, nil, nil }
};


static int
codec(char *path)
{
	char *arg[10];
	int i, narg, fd[2];

	narg = 0;
	arg[narg++] = path;
	arg[narg++] = "-s";
	arg[narg++] = "-v";
	arg[narg] = nil;

	if(dbg("codec"))
		fprint(2, "codec %s -s -v\n", path);

	if(pipe(fd) < 0){
		fprint(2, "pipe failed %r");
		return -1;
	}

	switch(fork()){
	case -1:
		fprint(2, "fork failed %r");
		close(fd[0]);
		close(fd[1]);
		return -1;
	case 0:
		dup(fd[0], 0);
		close(fd[0]);
		close(fd[1]);
		for(i = 3; i < 63; i++)
			close(i);
		setprio(getpid(), 16);
		exec(path, arg);
		sysfatal("exec %s failed %r", path);
	default:
		close(fd[0]);
		break;
	}
	return fd[1];
}

static int
stream(Sess *s, int conn, int net, char *path)
{
	char buf[8192];
	int got, n, len, metaint;

	metaint = icyhdr(s, conn);

	/* start a codec only on changes */
	if(strcmp(s->decpath, path) != 0){

		strcpy(s->decpath, path);
		if(s->dec != -1)
			close(s->dec);

		if(Outfile){
			if((s->dec = create(Outfile, OWRITE, 0644)) == -1)
				return Error;
		}
		else{
			if(Touch)
				return Finished;
			if((s->dec = codec(path)) == -1)
				return Error;
		}
	}


	if(metaint == -1){
		while(1){
			if((n = read(net, buf, sizeof(buf))) < 1){
				werrstr("end of stream");
				break;
			}
			if(write(s->dec, buf, n) != n)
				break;
		}
	}
	else{
		while(1){
			for(got = 0; got < metaint; got += n){
	
				len = sizeof(buf);
				if(metaint - got < len)
					len = metaint - got;
	
				if((n = read(net, buf, len)) < 1){
					werrstr("end of stream");
					goto done;
				}
				if(write(s->dec, buf, n) != n)
					goto done;
			}
		if(icyinband(net) < 0)
			break;
		}
	}

done:
	return Finished;
}

/*
 * Try to discover the content type if an http connection.
 * we only look at mimetype and the extension on the filename at
 * present.
 *
 * We could load the first kbyte of the stream and analyse
 * that but it has not been necessary (yet). I have tried to avoid this
 * as it would mean poluting the stream agnostic radio app with stream
 * specific header analysis code, which in the case of mp4 or ogg could
 * turn out to be very complex.
 *
 * I could envoke file(1) but I fear that would need to be taught about
 * more file formats to cope.
 */
static Mime *
mimetype(Sess *s, int conn, char *url)
{
	int n;
	Mime *m;
	char *a[32], *tail, *ext, buf[1024];

	*buf = 0;
	if(httphdr(s, conn, "contenttype", buf, sizeof(buf)) == 0){
		if(dbg("mime"))
			fprint(2, "mime raw %q\n", buf);

		/* some sites add other semicolon or whitespace delimited fields */
		n = getfields(buf, a, nelem(a), 1, ";, \t\n");
		if(n > 0){
			if(dbg("mime"))
				fprint(2, "mime %s\n", a[0]);
			for(m = Mimetypes; m->type; m++)
				if(cistrcmp(m->type, a[0]) == 0)
					return m;
		}
	}

	/*
	 * No ContenType, or an unrecognised one,
	 * try and match by filename extension (ugh!)
	 */
	if((ext = strrchr(url, '/')) != nil)
		if((ext = strchr(ext, '.')) != nil){
			for(m = Mimetypes; m->type; m++)
				if(cistrcmp(m->ext, ext+1) == 0)		/* +1 to step over '.' */
					return m;
		}

	if((tail = strrchr(url, '/')) == nil)
		tail = "?";
	werrstr("%s cannot discern contents", tail);
	return nil;
}

void
dumplist(Playlist *pl)
{
	Chunk *ch;

	fprint(2, "seqbase=%lld ", pl->seqbase);
	fprint(2, "type=%s ", (pl->chunked)? "chunked": "stream");
	fprint(2, "source=%s ", (pl->live)? "live": "on-demand");
	fprint(2, "content=%s ", (pl->isaudio)? "audio": "playlist");
	fprint(2, "\n");

	for(ch = pl->chunks; ch; ch = ch->next){
		fprint(2, " url %q\n", ch->url);
		fprint(2, "	title %q\n", ch->title);
		fprint(2, "	secs %g\n", ch->secs);
		fprint(2, "	bandwidth %d\n", ch->bw);
		if(ch->start == 0 && ch->end == 0)
			fprint(2, "	range <whole file>\n");
		else
			fprint(2, "	range %lld - %lld\n", ch->start, ch->end);
	}		
}

void
timestamp(long then, char *type)
{
	double sec;
	static double mean = 0;
	static double maxval = -999;
	static double minval = 999;

	sec = (double)(nsec()-then)/1e9;
	if(mean == 0)
		mean = sec;
	mean = (mean + sec)/2.0;
	if(sec > maxval)
		maxval = sec;
	if(sec < minval)
		minval = sec;

	fprint(2, "time min=%.3f ave=%.3f max=%.3f %.3f %s\n", sec, minval, mean, maxval, type);
}

void
showmeta(Sess *s, Chunk *ch)
{
	if(ch->title)
		if(s->title == nil || strcmp(s->title, ch->title) != 0){
			print("title %q\n", ch->title);
			free(s->title);
			s->title = ch->title;
			ch->title = nil;
		}

	if(ch->artist)
		if(s->artist == nil || strcmp(s->artist, ch->artist) != 0){
			print("artist %q\n", ch->artist);
			free(s->artist);
			s->artist = ch->artist;
			ch->artist = nil;
		}

}

/*
 * Play a given url.
 *
 * may need to call recursively if the url is a playlist.
 *
 * plays HLS chunk'ed streams, taking care not to duplicate
 * chunks already played.
 */
int
play(Sess *s, int conn, char *url, vlong start, vlong end)
{
	Mime *m;
	Chunk *ch;
	Playlist *pl;
	vlong seq, then;
	int fd, status;

	status = Error;

	then = nsec();
	if((fd = httpdial(s, conn, url, start, end)) < 0){
		werrstr("cannot connect, %r");
		return status;
	}

	if((m = mimetype(s, conn, url)) == nil){
		close(fd);
		return status;
	}

	if(dbg("time"))
		timestamp(then, m->type);

	if(m->codec && *m->codec){
		status = stream(s, conn, fd, m->codec);
		close(fd);
		return status;
	}

	pl = (*m->func)(url, fd);

	close(fd);

	if(pl == nil){
		werrstr("playlist empty");
		return status;
	}

	if(dbg("playlist"))
		dumplist(pl);

	if(s->seqno == 0)			/* first we have seen, also, maybe always zero */
		s->seqno = pl->seqbase;

	if(dbg("seq"))
		fprint(2, "seq newlist want=%lld base=%lld live=%d\n",
			s->seqno, pl->seqbase, pl->live);

	if(pl->isaudio){
		seq = pl->seqbase;
		ch = pl->chunks;

		for(; ch && seq < s->seqno; ch = ch->next){
			if(dbg("dupe"))
				fprint(2, "dupe want=%lld got=%lld\n", s->seqno, seq);
			seq++;
		}
		if(seq > s->seqno)
			if(dbg("loss"))
				fprint(2, "loss want=%lld got=%lld\n", s->seqno, seq);

		if(pl->chunked)
			s->seqno = seq;

		for(; ch; ch = ch->next){
			if(dbg("seq"))
				fprint(2, "seq audio got=%lld\n", s->seqno);
			showmeta(s, ch);
			status = play(s, conn+1, ch->url, ch->start, ch->end);
			if(status == Error)
				break;
			if(pl->chunked)
				s->seqno++;
		}
	}
	else{
		for(ch = pl->chunks; ch; ch = ch->next)
			do{
				if(dbg("seq"))
					fprint(2, "seq list\n");
				status = play(s, conn+1, ch->url, ch->start, ch->end);
				if(dbg("seq"))
					fprint(2, "seq ret=%d\n", status);
			}while(status == Retry);
	}

	if(dbg("seq"))
		fprint(2, "seq done status=%d live=%d\n", status, pl->live);

	if(! Touch && status == Finished && pl->live){
		if(! pl->chunked){
			/* we expect non-chunked streams to last a long time (days) */
			fprint(2, "%s: %s EOF on stream, retrying\n", argv0, url);
		}
		status = Retry;
	}

	freeplaylist(pl);
	return status;
}

static void
usage(void)
{
	fprint(2, "usage: %s [-n] url\n", argv0);
	fprint(2, " -n  narrowband mode, prefer low bitrate HLS streams\n");
	fprint(2, " -t  touch, refresh DNS but don't read audio\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	Sess sess, *s;
	int rc, n, i;
	char buf[128], err[ERRMAX];

	memset(&sess, 0, sizeof(Sess));

	s = &sess;
	Outfile = nil;
	ARGBEGIN{
	case 'o':
		Outfile = EARGF(usage());
		break;
	case 'd':
		Debug = EARGF(usage());
		break;
	case 'n':
		Narrowband = 1;
		break;
	case 't':
		Touch = 1;
		break;
	}ARGEND;

	quotefmtinstall();

	if(argc != 1)
		usage();

	s->dec = -1;
	s->webfs = "/mnt/web";

	/*
	 * we need the word "iTunes" here to force live365.com
	 * to return a valid Content-Type header field.
	 */
	s->agent = "iTunes (really: Plan9 IR4)";

	for(i = 0; i < Nconn; i++){
		snprint(buf, sizeof(buf), "%s/clone", s->webfs);
		if((s->net[i] = open(buf, ORDWR)) < 0)
			sysfatal("%s cannot open - %r\n", buf);

		if((n = read(s->net[i], buf, sizeof buf-1)) <= 0)
			sysfatal("%s cannot read - %r\n", buf);
		buf[n] = '\0';
		s->conid[i] = atoi(buf);
	}

	setprio(getpid(), 16);

	while((rc = play(s, 0, argv[0], 0, 0)) == Retry)
		continue;

	if(rc == Error){
		rerrstr(err, sizeof(err));
		print("error %q\n", err);
		exits(err);
	}

	exits(nil);
}
