#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include "dat.h"
#include "fns.h"

int
dbg(char *s)
{
	if(Debug != nil && strstr(Debug, s) != nil)
		return 1;
	return 0;
}

/*
 * Here we take care to read streams which have valid data on
 * their last line but are missing the last newline character.
 */
int
Bgetline(Biobuf *b, char *buf, int nbuf)
{
	Rune r;
	int i, rv;

	for(i = 0; i < nbuf - UTFmax - 1; ){
		rv = Bgetrune(b);

		if(rv == -1){
			if(i == 0)
				return -1;
			break;
		}
		if(rv == '\n')
			break;
		r = rv;
		i += runetochar(buf+i, &r);
	}
	if(i > 0 && buf[i-1] == '\r')
		buf[i-1] = 0;
	buf[i] = 0;
	return i;
}

int
setprio(int pid, int prio)
{
	int fd;
	char buf[128];

	snprint(buf, sizeof(buf), "/proc/%d/ctl", pid);
	if((fd = open(buf, OWRITE)) == -1)
		return -1;
	fprint(fd, "pri %d\n", prio);
	close(fd);
	return 0;
}

char *
skipws(char *p)
{
	while(isspace(*p))
		p++;
	return p;
}

void
dmphdrs(Sess *s, int conn)
{
	Dir *d;
	int n, m, i, fd;
	char buf[1024];

	snprint(buf, sizeof(buf), "%s/%d", s->webfs, s->conid[conn]);
	if((fd = open(buf, OREAD)) == -1)
		return;
	n = dirreadall(fd, &d);
	close(fd);

	for(i = 0; i < n; i++){
		snprint(buf, sizeof(buf), "%s/%d/%s", s->webfs, s->conid[conn], d[i].name);
		fprint(2, "	%-32s ", buf);
		if((fd = open(buf, OREAD)) != -1)
			if((m = read(fd, buf, sizeof(buf)-1)) > 0)
				write(2, buf, m);
		fprint(2, "\n");
		close(fd);
	}
	free(d);
}

int
httpredial(Sess *s, int conn, char *url, vlong start, vlong end, char *etag)
{
	int fd, tries;
	char buf[128], err[ERRMAX];

	tries = 0;
	if(dbg("http"))
		fprint(2, "http.dial %lld->%lld %q\n", start, end, url);
Retry:
	snprint(buf, sizeof buf, "%s/%d/ctl", s->webfs, s->conid[conn]);
	if((fd = open(buf, OWRITE)) < 0){
		werrstr("cannot open another HTTP session, %r");
		return -1;
	}
	fprint(fd, "url %s", url);
	if(s->agent && *s->agent)
		fprint(fd, "headers User-agent: %s", s->agent);
	fprint(fd, "headers icy-metadata: 1");
	if(etag && *etag)
		fprint(fd, "headers If-None-Match: %s", etag);
	if(start < end)
		fprint(fd, "headers Range: bytes=%lld-%llud", start, end);
	close(fd);

	snprint(buf, sizeof buf, "%s/%d/body", s->webfs, s->conid[conn]);
	fd = open(buf, OREAD);

	if(fd < 0){
		rerrstr(err, sizeof(err));
		if(dbg("http"))
			fprint(2, "http.dial failed err=[%s]\n", err);

		/*
		 * Wait whilst the socket closes, usually less than a second.
		 * We could use 2 sockets and ping pong between them but this
		 * is good enough for now.
		 */
		if(strstr(err, "client in use") != nil && tries++ < 10){
			sleep(300);
			goto Retry;
		}

		httperrstr(err);
		return -1;
	}

	if(dbg("dmphdrs"))
		dmphdrs(s, conn);
	return fd;
}

int
httpdial(Sess *s, int conn, char *url, vlong start, vlong end)
{
	return httpredial(s, conn, url, start, end, nil);
}


int
httphdr(Sess *s, int conn, char *field, char *val, int len)
{
	int n, fd;
	char buf[128];

	if(val && len > 0)
		*val = 0;

	snprint(buf, sizeof(buf), "%s/%d/%s", s->webfs, s->conid[conn], field);
	if((fd = open(buf, OREAD)) < 0)
		return -1;
	n = read(fd, val, len-1);
	close(fd);

	if(n < 0)
		return -1;
	val[n] = 0;
	return 0;
}

void *
emalloc(long n)
{
	void *p;

	if((p = mallocz(n, 1)) == nil)
		sysfatal("No memory, %r");
	return p;
}

char *
estrdup(char *s)
{
	if((s = strdup(s)) == nil)
		sysfatal("No memory, %r");
	return s;
}

char *
esmprint(char *fmt, ...)
{
	char *s;
	va_list ap;

	va_start(ap, fmt);
	s = vsmprint(fmt, ap);
	va_end(ap);

	if(s == nil)
		sysfatal("No memory, %r");
	return s;
}

char *
absurl(char *base, char *url)
{
	char **m, *methods[] = { "http://", "ftp://", nil };

	for(m = methods; *m; m++)
		if(strncmp(*m, url, strlen(*m)) == 0)
			return estrdup(url);

	return esmprint("%s/%s", base, url);
}

char *
baseurl(char *url)
{
	char *p, *base;

	base = estrdup(url);
	if((p = strstr(base, "://")) == nil)
		sysfatal("%s bad url\n", url);
	p += 3;
	if((p = strrchr(p, '/')) == nil)
		sysfatal("%s bad url\n", url);
	*p = 0;
	return base;
}

void
freeplaylist(Playlist *pl)
{
	Chunk *ch, *t;

	for(ch = pl->chunks; ch; ch = t){
		t = ch->next;
		free(ch->aux);
		free(ch->url);
		free(ch->title);
		free(ch);
	}
	free(pl->aux);
	free(pl);
}

/*
 * Addicted radio HLS chunks contain metadata of the form:
 *  title="North American Scum",artist="<BOM>LCD Soundsystem"
 *
 * Jazz paino magic prepends the track title with with "(#1 - 19/9999) " 
 */
void
parsetitle(Chunk *ch, char *buf)
{
	char *p;
	char *spc = "﻿";											/* zero length, non-breaking space */

	if(buf == nil)
		return;
	if(strstr(buf, "no desc") != nil)							/* The BBC, being annoying */
		return;

	if(strncmp(buf, "(#1 - ", 6) == 0)
		if((p = strstr(buf, "/9999) ")) != nil){				/* Jazz piano magic */
			ch->title = estrdup(p +7);
			return;
		}

	if(strstr(buf, "title=\"") != nil){							/* Addicted radio */
		ch->title = estrdup(buf);
		p = strstr(ch->title, "title=\"") +7;
		strcpy(ch->title, p);
		if((p = strchr(ch->title, '"')) != 0)
			*p = 0;
	
		ch->artist = estrdup(buf);
		p = strstr(ch->artist, "artist=\"") +8;
		if(memcmp(p, spc, sizeof(spc)-1) == 0)					/* strip prepended non-brk-space */
			p += sizeof(spc)-1;
		strcpy(ch->artist, p);
		if((p = strchr(ch->artist, '"')) != 0)
			*p = 0;
		return;
	}

	if(strstr(buf, " - ") != nil){								/* last chance */
		ch->title = estrdup(buf);
		p = strstr(ch->title, " - ");
		*p = 0;

		ch->artist = estrdup(buf);
		p = strstr(ch->artist, " - ") +3;
		strcpy(ch->artist, p);
		return;
	}

	ch->title = estrdup(buf);
}
