#include <u.h>
#include <libc.h>
#include <draw.h>
#include <html.h>
#include <fcall.h>
#include <thread.h>
#include <bio.h>
#include <9p.h>
#include <ctype.h>
#include "dat.h"
#include "fns.h"


enum {				/* state machine states */
	Lost,			/* searching for "Now on" */
	Nowon,			/* searching for time stamp */
	Gottime,		/* print next line */
	Finished		/* success */
};

typedef struct Growbuf Growbuf;
struct Growbuf {		/* growable buffer */
	int len;
	int alloc;
	char *buf;
};

/* for libhtml */
void *
emalloc(ulong n)
{
	void *p;

	if((p = mallocz(n, 1)) == nil)
		sysfatal("No memory, %r");
	return p;
}

/* for libhtml */
void *
erealloc(void *p, ulong n)
{
	void *q;

	if((q = realloc(p, n)) == nil)
		sysfatal("No memory, %r");
	return q;
}

Growbuf *
allocbuf(void)
{
	Growbuf *gb;

	gb = mallocz(sizeof(Growbuf), 1);
	if(gb == nil)
		sysfatal("No memory, %r\n");
	memset(gb, 0, sizeof(Growbuf));
	return gb;
}

void
growbuf(Growbuf *gb, char *buf, int len)
{
	if(gb->len + len >= gb->alloc){
		gb->alloc = gb->len + len + 1024;
		gb->buf = erealloc9p(gb->buf, gb->alloc);
	}
	memcpy(gb->buf + gb->len, buf, len);
	gb->len += len;
}

void
freebuf(Growbuf *gb)
{
	free(gb->buf);
	free(gb);
}

static int
lookcharset(char *s)
{
	char *meta, *emeta, *charset;

	meta = cistrstr(s, "<meta");
	if(meta == nil)
		return ISO_8859_1;
	for(emeta=meta; *emeta!='>' && *emeta!='\0'; emeta++)
		continue;
	charset = cistrstr(s, "charset=");
	if(charset == nil)
		return ISO_8859_1;
	charset += 8;
	if(*charset == '"')
		charset++;
	if(cistrncmp(charset, "utf-8", 5) || cistrncmp(charset, "utf8", 4))
		return UTF_8;
	return ISO_8859_1;
}

/*
 * returns true if the string is a
 * timestamp of the form:
 *    06:27 a.m.
 */
int
istimestamp(Rune *r)
{

	if(! isdigitrune(*r))
		return 0;
	while(isdigitrune(*r))
		r++;
	if(*r++ != ':')
		return 0;
	if(! isdigitrune(*r))
		return 0;
	while(isdigitrune(*r))
		r++;
	while(isspacerune(*r))
		r++;

	if(runestrcmp(r, L"a.m.") == 0)
		return 1;
	if(runestrstr(r, L"p.m.") == 0)
		return 1;
	return 0;
}

/*
 * Usual nasty heuristics to try and split
 * "artist - title" into two seperate fields.
 */
void
prnt(char *str)
{
	char *p;

	if((p = strstr(str, " - ")) != nil){
		*p = 0;
		print("artist %q\n", str);
		str = p +3;
	}
	print("title %q\n", str);
}

/*
 * Beware: this is very crufty, scraping from their
 * website using very crude heuristics. I have complained
 * asking them to projide json, xml, or even just old skool
 * shoutcast metadata but have heard nothing yet.
 * This will have to do for now.
 */
int
statemachine(int state, Fmt *f, Rune *str)
{
	while(isspacerune(*str))	/* skip leading whitespace */
		str++;
	if(*str == 0)				/* ignore blank lines */
		return state;

	switch(state){
	case Lost:
		if(runestrcmp(str, L"Now On") == 0)
			state = Nowon;
		break;
	case Nowon:
		if(istimestamp(str))
			state = Gottime;
		break;
	case Gottime:
		if(istimestamp(str))
			state = Finished;
		else
			fmtprint(f, "%S", str);
	}

	return state;
}

static int
parse(char *url, Growbuf *gb, int charset)
{
	Fmt f;
	Itext *it;
	int state;
	Rune *rsrc;
	Docinfo *di;
	Item *items, *i;

	if(charset == UnknownCharset)
		charset = lookcharset(gb->buf);

	di = nil;
	rsrc = runesmprint("%s", url);
	if((items = parsehtml((uchar *)gb->buf, gb->len, rsrc, TextHtml, charset, &di)) == nil){
		freedocinfo(di);
		free(rsrc);
		return 0;
	}


	fmtstrinit(&f);
	state = Lost;
	for(i = items; i; i = i->next){
		if(i->tag == Itexttag){
			it = (struct Itext *)i;
			state = statemachine(state, &f, it->s);
			if(state == Finished)
				break;
		}
	}
	prnt(fmtstrflush(&f));

	freeitems(items);
	freedocinfo(di);
	free(rsrc);
	return 1;
}


void
poll(Sess *s, char *url)
{
	int n, m, fd;
	long msec, len, got;
	Growbuf *gb;
	char buf[8192], etag[1024];

	*etag = 0;
	msec = 20*1000;
	for(;;){
		if((fd = httpredial(s, url, etag, sizeof(etag))) < 0){
			if(! notmodified())
				fprint(2, "%s: %s httpredial failed, %r\n", argv0, url);
			goto again;
		}

		if(httphdr(s, "contentlength", buf, sizeof(buf)) < 0){
			fprint(2, "%s: cannot get Content-Length\n", argv0);
			close(fd);
			goto again;
		}
		len = atol(buf);

		gb = allocbuf();
		for(got = 0; got < len; got += m){
			n = sizeof(buf);
			if(n > len-got)
				n = len-got;
			if((m = read(fd, buf, n)) <= 0)
				break;
			growbuf(gb, buf, m);
		}
		close(fd);
	
		if(gb->buf == nil || *gb->buf == 0){
			fprint(2, "%s: unexpected EOF, %r\n", argv0);
			freebuf(gb);
			goto again;
		}
		parse(url, gb, UnknownCharset);
		freebuf(gb);

again:
		sleep(msec/2 + lnrand(msec));
	}
}
