#include <u.h>
#include <libc.h>
#include <draw.h>
#include <keyboard.h>
#include <event.h>
#include <ctype.h>
#include <bio.h>

typedef struct Metadata Metadata;
struct Metadata {
	char *str;
	int dirty;
	Rectangle r;
};

enum {			/* These are dissplayed in order so later items have priority */
	Mstation,
	Mgenre,
	Mshow,
	Martist,
	Mtitle,
	Mrate,
	Mfreq,
	Mchans,
	Menc,
	Memph,

	Nmeta		/* must be last */
};


struct {
	char *s;
	int n;
} Map[] = {
	{ "station",	Mstation },		/* usually from shoutcast http headers */
	{ "genre",		Mgenre },

	{ "show",		Mshow },		/* from inband metadata or external helper program */
	{ "artist",		Martist },
	{ "title",		Mtitle },

	{ "enc",		Menc },			/* from audio decoder */
	{ "chan",		Mchans },
	{ "freq",		Mfreq },
	{ "rate",		Mrate },
	{ "emph",		Memph },
};


typedef struct Channel Channel;
struct Channel {
	char *audiourl;		/* audio stream, .PLS, or .M4U file URL */
	char *metaurl;		/* optional metadata url */
	char *metaprog;		/* optional metadata class type */
	char *label;		/* what to display in the title field */
	long atime;			/* last accessed, i.e. DNS last refreshed */
};

typedef struct Volume Volume;
struct Volume {
	int l;
	int r;
	int chans;
};

enum {						/* GUI layout, widths in chars */
	Wcol1 = 16, 			/* width of collum 1 */
	Wstation = 32,			/* width of station name */
	Wgenre = 32,			/* width of genre */
	Wtitle = 50,			/* width of title, artist, source */

	Horz = 1, 				/* gap between vertical elements */
	Vert = 0 				/* gap between horizontal elements */
};

int Poke;					/* keep pokeing dns to make channel switching faster */
Volume Vol;					/* current volume level */
Metadata Meta[Nmeta];		/* metadata read from radio player */
QLock Lck;					/* lock on metadata */
int Current = -1;			/* currently selected channel */
Channel *Chans;				/* array of channels */
int Nchans;					/* size of above array */
int Audproc;				/* pid of current radio player */
int Metaproc;				/* pid or current metadata parser */
Font *Face;					/* current font */
Image *Background;			/* background colour */
char Chanlist[128];			/* filename of channel list */
int Debug;					/* chatter */
int Fixname;				/* don't update station name from metadata */


int chans;					/* cached data, for compression ratio calculation */
double rate, freq;			/* cached data, for compression ratio calculation */

enum {
	PAD = 5,
	MARGIN = 8,
	Chunk = 16,

	But1 = 1,
	But2 = 2,
	But3 = 4,
};

void*
erealloc(void *v, ulong n)
{
	v = realloc(v, n);
	if(v == nil)
		sysfatal("out of memory reallocating %lud", n);
	return v;
}

void*
emalloc(ulong n)
{
	void *v;

	v = malloc(n);
	if(v == nil)
		sysfatal("out of memory allocating %lud", n);
	memset(v, 0, n);
	return v;
}

char*
estrdup(char *s)
{
	int l;
	char *t;

	if (s == nil)
		return nil;
	l = strlen(s)+1;
	t = emalloc(l);
	memcpy(t, s, l);

	return t;
}

char *
trim(char *s)
{
	char *p;

	for(p = strrchr(s, 0) -1; p >= s && isspace(*p); p--)
		*p = 0;
	for(p = s; isspace(*p); p++)
		continue;
	return p;
}

void
readlist(void)
{
	int n, m;
	Biobuf *bp;
	char *line, *a[8];

	if((bp = Bopen(Chanlist, OREAD)) == nil)
		return;

	m = 0;
	Nchans = 0;
	while((line = Brdline(bp, '\n')) != nil){
		line[Blinelen(bp)-1] = 0;
		n = tokenize(line, a, nelem(a));
		if(n > 0 && a[0][0] == '#')
			continue;
		if(Nchans >= m){
			m += Chunk;
			Chans = erealloc(Chans, (m+Chunk) * sizeof(Channel));
		}
		memset(&Chans[Nchans], 0, sizeof(Chans[Nchans]));
		switch(n){
		case 4:
			Chans[Nchans].metaurl = estrdup(a[3]);
		case 3:
			Chans[Nchans].metaprog = estrdup(a[2]);
		case 2:
			Chans[Nchans].audiourl = estrdup(a[1]);
		case 1:
			Chans[Nchans].label = estrdup(a[0]);
			break;
		case 0:
			Chans[Nchans].label = estrdup("");
			break;
		}
		Nchans++;
	}
	Bterm(bp);
}

void
readvolume(void)
{
	int n;
	Biobuf *bp;
	char *line, *a[4];

	if((bp = Bopen("/dev/volume", OREAD)) == nil)
		return;
	while((line = Brdline(bp, '\n')) != nil){
		line[Blinelen(bp)-1] = 0;
		n = tokenize(line, a, nelem(a));
		if(n < 2)
			continue;
		if(strcmp(a[0], "audio") == 0)
			switch(n){
			case 2:
				Vol.r = Vol.l = atoi(a[1]);
				Vol.chans = 1;
				break;
			case 3:
				Vol.l = atoi(a[1]);
				Vol.r = atoi(a[2]);
				Vol.chans = 2;
				break;
			default:
				break;
			}
	}
	Bterm(bp);
}

void
set(int fld, char *val)
{
	qlock(&Lck);
	free(Meta[fld].str);
	Meta[fld].str = estrdup(trim(val));
	Meta[fld].dirty = 1;
	qunlock(&Lck);
}


void
updfreq(int m, double freq)
{
	if(freq > 160 && freq < 200)
		set(m, "quadruple");
	if(freq > 80 && freq < 100)
		set(m, "double");
	if(freq > 40 && freq < 50)
		set(m, "full");
	if(freq > 30 && freq < 39)
		set(m, "reduced");
	if(freq > 20 && freq < 25)
		set(m, "half");
	if(freq > 10 && freq < 13)
		set(m, "quarter");
}

void
updrate(int m, int chans, int depth, double freq, double rate)
{
	char buf[32];
	double ratio;

	if(rate == 0)
		return;
	ratio = (freq*1000*chans*depth*8) / (rate*1024) + 0.5;
	snprint(buf, sizeof(buf), "%.0f:1", ratio);
	set(m, buf);
}

void
readmeta(int fd)
{
	int i, m;
	Biobuf in;
	char *line, *a[6];

	Binit(&in, fd, OREAD);
	while((line = Brdline(&in, '\n')) != nil){
		line[Blinelen(&in)-1] = 0;
		if(tokenize(line, a, nelem(a)) < 2)
			continue;

		if(strcmp(a[0], "error") == 0)
			set(Mtitle, a[1]);

		for(i = 0; i < nelem(Map); i++)
			if(strcmp(a[0], Map[i].s) == 0){
				m = Map[i].n;

				if(Fixname && m == Mstation)
					continue;

				set(m, a[1]);
				switch(m){
				case Mfreq:
					freq = atof(a[1]);
					updfreq(Mfreq, freq);
					updrate(Mrate, chans, 2, freq, rate);
					break;
				case Mrate:
					rate = atof(a[1]);
					updrate(Mrate, chans, 2, freq, rate);
					break;
				case Mchans:
					if(cistrcmp(a[1], "mono") == 0)
						chans = 1;
					else
					if(cistrcmp(a[1], "stereo") == 0 ||
					   cistrcmp(a[1], "joint") == 0 ||
					   cistrcmp(a[1], "dual") == 0)
						chans = 2;
					else
						chans = atoi(a[1]);
					updrate(Mrate, chans, 2, freq, rate);
					break;
				default:
					break;
				}
			}
	}
	Bterm(&in);
}

void
drawmeta(int i)
{
	draw(screen, Meta[i].r, Background, nil, ZP);
	if(Meta[i].str)
		_string(screen, addpt(Meta[i].r.min, Pt(2,0)), display->black, ZP,
			Face, Meta[i].str, nil, strlen(Meta[i].str), 
			Meta[i].r, nil, ZP, SoverD);
	if(Debug)
		border(screen, Meta[i].r, 1, display->black, ZP);	
	Meta[i].dirty = 0;
}

Rectangle 
field(Point p, Point fs, int n)
{
	if(n < 0)
		return Rpt(addpt(screen->r.min, p), addpt(screen->r.min, Pt(screen->r.max.x- (fs.x*Horz), p.y+fs.y)));
	else
		return Rpt(addpt(screen->r.min, p), addpt(screen->r.min, Pt(p.x+fs.x*n, p.y+fs.y)));
}

int
geometry(void)
{
	Point p1, p2, p3;
	static Rectangle scr;
	Point fs = stringsize(Face, "j");

	if(eqrect(scr, screen->r))
		return 0;
	scr = screen->r;

	p1 = Pt((fs.x*Horz), (fs.y*Vert));
	Meta[Mchans].r = field(p1, fs, Wcol1);
	p2 = Pt(p1.x+Dx(Meta[Mchans].r), p1.y);
	Meta[Mstation].r = field(p2, fs, Wstation);
	p3 = Pt(p2.x+Dx(Meta[Mstation].r), p1.y);
	Meta[Mgenre].r = field(p3, fs, -1);

	p1 = Pt(p1.x, p1.y+Dy(Meta[Mchans].r)+(fs.y*Vert)/2);
	Meta[Mfreq].r = field(p1, fs, Wcol1);

	p1 = Pt(p1.x, p1.y+Dy(Meta[Mfreq].r)+(fs.y*Vert)/2);
	Meta[Menc].r = field(p1, fs, Wcol1);
	p2 = Pt(p1.x+Dx(Meta[Menc].r), p1.y-(fs.y*Vert));
	Meta[Mshow].r = field(p2, fs, -Wtitle);

	p1 = Pt(p1.x, p1.y+Dy(Meta[Menc].r)+(fs.y*Vert)/2);
	Meta[Mrate].r = field(p1, fs, Wcol1);
	p2 = Pt(p1.x+Dx(Meta[Mrate].r), p1.y);
	Meta[Martist].r = field(p2, fs, -Wtitle);

	p1 = Pt(p1.x, p1.y+Dy(Meta[Mrate].r)+(fs.y*Vert)/2);
	Meta[Memph].r = field(p1, fs, Wcol1);
	p2 = Pt(p1.x+Dx(Meta[Memph].r), p1.y);
	Meta[Mtitle].r = field(p2, fs, -Wtitle);

	return 1;
}

void
clr(char *s)
{
	freq = 0;
	rate = 0;
	chans = 0;
	set(Martist, "⋯");
	set(Mchans, "⋯");
	set(Menc, "⋯");
	set(Mfreq, "⋯");
	set(Mgenre, "⋯");
	set(Mshow, "⋯");
	set(Mrate, "⋯");
	set(Mtitle, "⋯");
	set(Memph, "");
	set(Mstation, s);
}

void
redraw(Image *screen, int all)
{
	int i;

	all |= geometry();
	if(all)
		draw(screen, screen->r, Background, nil, ZP);

	qlock(&Lck);
	for(i = 0; i < Nmeta; i++)
		if(all || Meta[i].dirty)
			drawmeta(i);
	qunlock(&Lck);
}

void
eresized(int new)
{
	if(new && getwindow(display, Refmesg) < 0)
		fprint(2,"can't reattach to Chansdow");
	geometry();
	redraw(screen, 1);
}

int
forkproc(char *path, char *prog, ...)
{
	va_list ap;
	int n, pid, fd[2];
	char *args[64], *full;

	if(pipe(fd) < 0)
		sysfatal("pipe failed %r");

	switch((pid = rfork(RFPROC|RFNOWAIT|RFNAMEG|RFNOTEG|RFENVG|RFFDG))){
	case -1:
		sysfatal("fork failed %r");
		return -1;
	case 0:
		n = 0;
		args[n++] = prog;
		va_start(ap, prog);
		while((args[n++] = va_arg(ap, char *)) != nil)
			continue;
		va_end(ap);
		args[n] = nil;

		dup(fd[0], 1);
		close(fd[0]);
		close(fd[1]);
		sleep(10);				/* just allow a little time for previous app to exit */
		full = smprint("%s/%s", path, prog);
		exec(full, args);
		sysfatal("%s exec failed %r", full);
	default:
		break;
	}

	switch(rfork(RFPROC|RFMEM|RFNOWAIT|RFFDG)){
	case -1:
		sysfatal("fork failed %r\n");
		break;
	case 0:				/* child */
		close(fd[0]);
		readmeta(fd[1]);
		close(fd[1]);
		exits(nil);
	default:
		close(fd[0]);
		close(fd[1]);
		break;
	}
	return pid;
}

char *
labgen(int i)
{
	if(i < 0 || i >= Nchans)
		return nil;
	return Chans[i].label;
}

void
newchan(int idx)
{
	/*
	 * allow spaces and titles to be selected
	 * so we can scroll with arrow keys
	 */
	Current = idx;

	if(Chans[idx].audiourl == nil)
		return;

	clr(Chans[idx].label);

	if(Audproc > 0)
		postnote(PNGROUP, Audproc, "kill");
	if(Metaproc > 0)
		postnote(PNGROUP, Metaproc, "kill");

	Audproc = forkproc("/bin/audio", "radio", Chans[idx].audiourl, nil);
	if(Chans[idx].metaurl && Chans[idx].metaprog)
		Metaproc = forkproc("/bin/audio/radiometa", Chans[idx].metaprog, Chans[idx].metaurl, nil);
	Chans[idx].atime = time(0);
}

int
runproc(char *path, char *prog, ...)
{
	va_list ap;
	int n, pid;
	char *args[64], *full;

	switch((pid = rfork(RFPROC|RFNAMEG|RFNOTEG|RFENVG|RFFDG))){
	case -1:
		sysfatal("fork failed %r");
		return -1;
	case 0:
		n = 0;
		args[n++] = prog;
		va_start(ap, prog);
		while((args[n++] = va_arg(ap, char *)) != nil)
			continue;
		va_end(ap);
		args[n] = nil;

		close(0);
		close(1);
		full = smprint("%s/%s", path, prog);
		exec(full, args);
		sysfatal("%s exec failed %r", full);
	default:
		break;
	}
	return pid;
}

void
pokedns(void)
{
	int i, p, pid;
	Waitmsg *w;
	long now, thresh;
	static int idx;
	static long last;

	now = time(nil);
	if(now - last < 2)						/* look every 2 secconds at most */
		return;
	last = now;

	thresh = 10*60 + lnrand(10*60);		/* poke dns every 10 - 20 mins */
	for(i = 0; i < Nchans; i++){
		if(Chans[idx].audiourl == nil){
			idx = (idx+1) % Nchans;
			continue;
		}
		if(now - Chans[idx].atime < thresh){
			idx = (idx+1) % Nchans;
			continue;
		}

		pid = runproc("/bin/audio", "radio", "-t", Chans[idx].audiourl, nil);
		while((w = wait()) != nil){
			if(w->msg == nil || *w->msg == 0){
				Chans[idx].atime = now;
				fprint(2, "success %lud\n", w->time[2]);
			}
			p = w->pid;
			free(w);
			if(p == pid)
				break;
		}
		idx = (idx+1) % Nchans;
		break;
	}
}

void
chansel(Mouse *m)
{
	static Menu men = { .gen = labgen };
	int hit;

	men.lasthit = Current;
	hit = emenuhit(3, m, &men);
	if(hit == -1 || hit == Current)
		return;
	newchan(hit);

}

void
bumpchan(int incr)
{
	if(Current + incr < 0 || Current >= Nchans)
		return;
	newchan(Current + incr);
}

void
ctrl(Mouse *m)
{
	int fd, hit;
	static char *item[] = { "Vol+", "Vol-", nil };
	static Menu men = { .item = item };

	hit = emenuhit(2, m, &men);
	if(hit == -1)
		return;

	switch(hit){
	case 0:
		Vol.l += 5;
		Vol.r += 5;
		break;
	case 1:
		Vol.l -= 5;
		Vol.r -= 5;
		break;
	default:
		return;
	}

	if(Vol.l < 0 || Vol.r < 0)
		Vol.l = Vol.r = 0;
	if(Vol.l > 100 || Vol.r > 100)
		Vol.l = Vol.r = 100;

	if((fd = open("/dev/volume", OWRITE)) == -1){
		fprint(2, "%s: /dev/volume cannot open\n", argv0);
		return;
	}
	switch(Vol.chans){
	case 1:
		fprint(fd, "audio %d\n", (Vol.l+Vol.r)/2);
		break;
	case 2:
		fprint(fd, "audio %d %d\n", Vol.l, Vol.r);
		break;
	default:
		break;
	}
	close(fd);
}
	

void
usage(void)
{
	fprint(2, "usage: %s [-p] [-f font] [-s] [-n]\n", argv0);
	exits("usage");
}

void
main(int argc, char **argv)
{
	Event e;
	int Etimer;
	char *home, *fontname;

	fontname = "/lib/font/bit/lucidasans/unicode.8.font";
	ARGBEGIN{
	case 'p':
		Poke++;
		break;
	case 'd':
		Debug++;
		break;
	case 'f':
		fontname = EARGF(usage());
		break;
	case 'n':
		Fixname = 1;
		break;
	default:
		usage();
	}ARGEND

	if(argc)
		usage();

	if((home = getenv("home")) == nil)
		sysfatal("$home not set\n");
	snprint(Chanlist, sizeof(Chanlist), "/%s/lib/radio", home);

	initdraw(0, 0, "Chanswatch");
	Background = allocimagemix(display, DPalebluegreen, DWhite);
	if(Background == nil)
		sysfatal("allocimagemix: %r");
	if((Face = openfont(display, fontname)) == nil)
		sysfatal("%s not found", fontname);

	clr("Nothing slected");

	readvolume();
	readlist();
	redraw(screen, 1);
	einit(Emouse|Ekeyboard);
	Etimer = etimer(0, 2500);

	for(;;){
		switch(eread(Emouse|Ekeyboard|Etimer, &e)){
		case Ekeyboard:
			switch(e.kbdc){
			case 0x7F:
				if(Audproc > 0)
					postnote(PNGROUP, Audproc, "kill");
				if(Metaproc > 0)
					postnote(PNGROUP, Metaproc, "kill");
				exits(0);
				break;
			case Kup:
				bumpchan(-1);
				break;
			case Kdown:
				bumpchan(1);
				break;
			default:
				break;
			}
			break;
		case Emouse:
			if(e.mouse.buttons == But2)
				ctrl(&e.mouse);
			if(e.mouse.buttons == But3)
				chansel(&e.mouse);
			/* fall through  */
		default:
			redraw(screen, 0);
			if(Poke)
				pokedns();
			break;
		}
	}
}
